/***************************************************************************
 * <p>Title: Presence Detection</p>
 * <p>Description: Projects In Computer Science</p>
 * <p>Copyright: Copyright (c) 2007</p>
 * <p>Company: Columbia University; all rights reserved</p>
 *
 * @author Denis Abramov (daa82@columbia.edu)
 * @version 1.0
 *
 * Description:
 * This is an algo that uses hue, saturation and difference to find
 * skin and thereby detect human presence
 ***************************************************************************/

package com.presencedetection.algos;

import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.awt.Color;
import com.presencedetection.filters.MedianFilter;
import java.awt.image.DataBuffer;
import com.presencedetection.filters.DilationFilter;
import java.util.Arrays;

public class ColorDetectionAlgo implements PDAlgoIntf {


  public ColorDetectionAlgo() {
  }

  /**************************************************************************
   * Get the IHS image
   *
   * I = (logopp(red)+logopp(blue)+logopp(green))/3;
   * Rg = logopp(red)-I;
   * By = logopp(blue)-(I+logopp(red))/2;
   *
   *************************************************************************/
  private ImageData getIRB(BufferedImage image) {
     image = BackgroundSubtractionAlgo.removeBackgroundColor(image, false);

     ImageData i_array = new ImageData(image.getHeight(), image.getWidth());
     ImageData rg_array = new ImageData(image.getHeight(), image.getWidth());
     ImageData by_array = new ImageData(image.getHeight(), image.getWidth());

     for (int y = 0; y < image.getHeight(); y++) {
       for (int x = 0; x < image.getWidth(); x++) {
        Color c = new Color(image.getRGB(x, y));
        int I = (logop(c.getRed()) + logop(c.getBlue()) + logop(c.getGreen())) / 3;
        int Rg = logop(c.getRed()) - I;
        int By = logop(c.getBlue()) - (I + logop(c.getRed())) / 2;

        i_array.set(x, y, I);
        rg_array.set(x, y, Rg);
        by_array.set(x, y, By);
       }
     }

     double scale = (image.getWidth() + image.getHeight())/320;
     int scalei = (int) Math.round(scale);
     if (scalei == 0)
       scalei = 1;

     MedianFilter f = new MedianFilter(i_array);
     ImageData i_filtered_array = f.applyFilter(2 * 4 * scalei);

     ImageData MAD = i_array.subtract(i_filtered_array);
     MAD = MAD.abs();

     f = new MedianFilter(MAD);
     MAD = f.applyFilter(2 * 6 * scalei);

     ImageData hue = getHue(rg_array, by_array);
     ImageData saturation = getSaturation(rg_array, by_array);

     ImageData skin_texture = getSkinTextureRegions(MAD, hue, saturation);
     DilationFilter fd = new DilationFilter(skin_texture);
     skin_texture = fd.applyFilter();

     skin_texture = shrinkSkinTexture(skin_texture, hue, saturation);
     return skin_texture;
  }

  /***************************************************************************
   * shrink the area of the skin texture
   *
   **************************************************************************/
  private ImageData shrinkSkinTexture(ImageData skin, ImageData hue, ImageData saturation) {
    ImageData image = new ImageData(skin.getHeight(), skin.getWidth());
    for (int y = 0; y < image.getHeight(); y++) {
      for (int x = 0; x < image.getWidth(); x++) {
        double s = skin.get(x,y);
        double h = hue.get(x,y);
        double sa = saturation.get(x,y);

        if (skin.get(x, y) == 255 && 110 <= hue.get(x, y) &&
            hue.get(x, y) <= 180 && 0 <= saturation.get(x, y) &&
            saturation.get(x, y) <= 130)
          skin.set(x, y, 255);
        else
          skin.set(x, y, 0);
      }
    }
    return skin;
  }

  /***************************************************************************
   * Find out where the skin texture areas are
   *
   **************************************************************************/
  private ImageData getSkinTextureRegions(ImageData mad, ImageData hue, ImageData saturation) {
    ImageData image = new ImageData(mad.getHeight(), mad.getWidth());
    for (int y = 0; y < image.getHeight(); y++) {
      for (int x = 0; x < image.getWidth(); x++) {

        double m = mad.get(x,y);
        double h = hue.get(x,y);
        double s = saturation.get(x,y);
        if (mad.get(x, y) < 4.5 && 120 < hue.get(x, y) && hue.get(x, y) < 160 &&
            10 < saturation.get(x, y) && saturation.get(x, y) < 60)
          image.set(x, y, 255);

        else if (mad.get(x, y) < 4.5 && 150 < hue.get(x, y) &&
                 hue.get(x, y) < 180 &&
                 saturation.get(x, y) > 20 && saturation.get(x, y) < 80)
          image.set(x, y, 255);

        else
          image.set(x, y, 0);
      }
    }

    return image;
  }

  /************************************************************************
   * Calculate the hue
   *
   ************************************************************************/
  private ImageData getHue(ImageData rg_array, ImageData by_array) {
    ImageData image = new ImageData(rg_array.getHeight(), rg_array.getWidth());

    for (int y = 0; y < image.getHeight(); y++) {
      for (int x = 0; x < image.getWidth(); x++) {
        double v = Math.atan2(rg_array.get(x, y), by_array.get(x, y)) * (180d / Math.PI);
        image.set(x, y, v);
      }
    }

    return image;
  }

  /**************************************************************************
   * Calculate the saturation
   *
   *************************************************************************/
  private ImageData getSaturation(ImageData rg_array, ImageData by_array) {
    ImageData image = new ImageData(rg_array.getHeight(), rg_array.getWidth());

   for (int y = 0; y < image.getHeight(); y++) {
     for (int x = 0; x < image.getWidth(); x++) {
       double v = Math.sqrt((rg_array.get(x, y) * rg_array.get(x, y) + by_array.get(x, y) * by_array.get(x, y)));
       image.set(x, y, v);
     }
   }

   return image;
 }

 /***************************************************************************
  * normalize the colors
  *
  **************************************************************************/

 private int logop(int v) {
      return (int) (105 * Math.log(v+1));
  }


  /*****************************************************
   * analyze
   *
   * @param image BufferedImage
   * @return PDResults
   *****************************************************/
  public PDResults analyze(BufferedImage image) {
    ImageData i = getIRB(image);
    return new PDResults(i.isPresenceDetected(100), 1, i);
  }

  /**
   * isRunning
   *
   * @return boolean
   */
  public boolean isRunning() {
    return true;
  }

  /**
   * start
   *
   * @return boolean
   */
  public boolean start() {
    return false;
  }

}
